           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                June 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CMS = Infra:

(*****************************************************************

         1. Metric space (MS)
         2. Complete metric space (CMS)
         3. Contracting map
         4. Banach Theorem

 *****************************************************************)

axclass ms0 < type

consts 
  distance    :: "'a::ms0 * 'a::ms0 => real"  (* Distance function *)

consts
  cauchy     :: "'a::ms0 infinite_seq => bool"
  convergeTo :: "'a::ms0 infinite_seq => 'a::ms0 => bool"  
                ("_ convergeTo _" [55,55] 55)

  limit      :: "'a::ms0 infinite_seq => 'a::ms0"

defs
  cauchy_def :"cauchy xs == ALL delta::real. 0 < delta --> (EX n. ALL i j. 
                    ((n <= i & n <= j) --> (distance(xs i, xs j) < delta)))"
  convergeTo_def :"xs convergeTo x == ALL eps::real. 0 < eps --> (EX n. ALL m.
                    n <= m --> distance(x, xs m) < eps)"

  limit_def      :"limit xs == THE x. xs convergeTo x"

(********************************************************** 
                    metric space (MS)
 **********************************************************)

axclass
  ms < ms0
  positive_ms           : 
    "ALL (x::'a::ms0) (y::'a::ms0). 0 <= distance(x,y)"
  diagonal_ms           : 
    "ALL (x::'a::ms0) (y::'a::ms0). (distance(x,y) = 0) = (x = y)"
  symmetry_ms           : 
    "ALL (x::'a::ms0) (y::'a::ms0). distance(x,y) = distance(y,x)"
  triangle_inequality_ms: 
    "ALL (x::'a::ms0) (y::'a::ms0) (z::'a::ms0). 
       distance(x,z) <= distance(x,y)+distance(y,z)"

consts
  non_expanding     :: "('a::ms0 => 'b::ms0) => bool"
  contraction       :: "('a::ms0 => 'b::ms0) => bool"
  contraction_alpha :: "('a::ms0 => 'b::ms0) => real => bool"
  map_alpha         :: "('a::ms0 => 'b::ms0) => real => bool"

defs
  non_expanding_def    : 
     "non_expanding f == map_alpha f 1"
  contraction_def      : 
     "contraction f == EX (alpha::real). contraction_alpha f alpha"
  contraction_alpha_def: 
     "contraction_alpha f alpha == (alpha < 1 & map_alpha f alpha)"
  map_alpha_def: 
     "map_alpha f alpha == (0 <= alpha &
              (ALL x y. distance(f x,f y) <= (alpha * distance(x,y))))"

(********************************************************** 
                       MS lemmas 
 **********************************************************)

lemma same_pnt_zero: "ALL x::'a::ms. distance(x,x)=0"
apply (simp add: diagonal_ms)
done

lemma diff_pnt_pos:
  "ALL (x::'a::ms) (y::'a::ms). x ~= y --> 0 < distance(x,y)"
apply (intro allI impI)
apply (case_tac "distance (x, y) = 0")
apply (simp add: diagonal_ms)
apply (simp add: positive_ms real_less_le)
done


lemma diff_pnt_pos_inv:
  "ALL (x::'a::ms) (y::'a::ms). 0 < distance(x,y) --> x ~= y"
apply (intro allI impI)

apply (case_tac "x = y")
apply (simp add: same_pnt_zero)
apply (simp)
done

(*** alpha contraction --> contraction ***)

lemma contraction_alpha_contraction: 
  "ALL f alpha. contraction_alpha f alpha --> contraction f"
apply (auto simp add: contraction_def)
done

(*** contraction --> non expanding ***)

lemma contraction_non_expanding: 
  "ALL (f::'a::ms => 'b::ms). contraction f --> non_expanding f"
apply (intro allI impI)
apply (simp add: contraction_def contraction_alpha_def map_alpha_def
                 non_expanding_def)
apply (erule exE)
apply (intro allI)
apply (erule conjE)+
apply (drule_tac x="x" in spec)
apply (drule_tac x="y" in spec)

apply (insert ALL_real_mult_right_mono)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="1" in spec)
apply (drule_tac x="distance (x, y)" in spec)
apply (simp add: positive_ms)
done

(*** The limit is unique ***)

lemma unique_convergence: 
  "ALL (xs::'a::ms infinite_seq) (x::'a::ms) (y::'a::ms). 
     xs convergeTo x & xs convergeTo y --> x=y"
apply (intro allI impI)

apply (case_tac "x = y")
apply (simp)

apply (simp add: convergeTo_def)
apply (erule conjE)+

apply (drule_tac x="distance (x, y) /2" in spec)
apply (drule_tac x="distance (x, y) /2" in spec)
apply (simp add: diff_pnt_pos)
apply (erule exE)+
apply (rename_tac xs x y n m)
apply (drule_tac x="max n m" in spec)
apply (drule_tac x="max n m" in spec)
apply (simp add: le_maxI1 le_maxI2)

apply (insert triangle_inequality_ms)
apply (drule_tac x="x" in spec)
apply (drule_tac x="xs (max n m)" in spec)
apply (drule_tac x="y" in spec)

apply (simp add: symmetry_ms)
done

(*** Convergence implies cauchy seq***)

lemma convergece_cauchy: 
  "ALL (xs::'a::ms infinite_seq) (x::'a::ms).
       xs convergeTo x --> cauchy xs"
apply (intro allI impI)

apply (simp add: cauchy_def)
apply (intro allI impI)

apply (simp add: convergeTo_def)
apply (drule_tac x="delta/2" in spec)
apply (simp)
apply (erule exE)

apply (rule_tac x="n" in exI)
apply (intro allI impI)

apply (frule_tac x="i" in spec)
apply (drule_tac x="j" in spec)
apply (simp)

apply (insert triangle_inequality_ms)
apply (drule_tac x="xs i" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="xs j" in spec)
apply (simp add: symmetry_ms)
done

(********************************************************** 
                Complete metric space (CMS)
 **********************************************************)
axclass
  cms < ms

  comp_metirc_space: "ALL xs::'a::ms infinite_seq. 
                          cauchy xs --> (EX x. xs convergeTo x)"

(********************************************************** 
                     CMS lemmas
 **********************************************************)

lemma the_limit:
  "ALL (xs::'a::cms infinite_seq).
      cauchy xs --> xs convergeTo limit xs"
apply (intro allI impI)
apply (insert comp_metirc_space)
apply (drule_tac x="xs" in spec)
apply (simp)
apply (erule exE)

apply (insert unique_convergence)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="x" in spec)

apply (simp add: limit_def)
apply (insert ALL_theI)
apply (drule_tac x="%x. xs convergeTo x" in spec)
apply (auto)
done


(********************************************************** 
                     Banach Theory
 **********************************************************)

(*** step 1 (cauchy) ***)

lemma Banach_lm_contraction:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (n::nat).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i))) --> 
       distance(xs n, f(xs n)) <= alpha^n * distance(xs 0, f(xs 0))"
apply (intro allI)
apply (induct_tac n)
apply (simp)
apply (intro impI)
apply (simp)
apply (erule conjE)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (elim conjE)
apply (drule_tac x="0" in spec)    (* for convenienec *)

apply (drule_tac x="xs na" in spec)
apply (drule_tac x="f (xs na)" in spec)

apply (insert ALL_real_mult_left_mono)
apply (drule_tac x="distance (xs na, f(xs na))" in spec)
apply (drule_tac x="alpha ^ na * distance (xs 0, f(xs 0))" in spec)
apply (drule_tac x="alpha" in spec)
apply (simp)
done

lemma Banach_lm_triangle:
  "ALL (xs::'a::cms infinite_seq) (r::nat) (n::nat).
     distance(xs r, xs (r+n)) <= 
         prog_sum0 n (%m::nat. distance(xs (r+m-(1::nat)), xs (r+m)))"
apply (intro allI)
apply (induct_tac n)
apply (simp add: same_pnt_zero)

apply (simp)
apply (insert triangle_inequality_ms)
apply (drule_tac x="xs r" in spec)
apply (drule_tac x="xs (r + na)" in spec)
apply (drule_tac x="xs (Suc (r + na))" in spec)
apply (simp)
done

lemma Banach_lm_geo_prog_sum:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (r::nat) (n::nat).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i))) --> 
        prog_sum0 n (%m::nat. distance(xs (r+m-(1::nat)), xs (r+m)))
        <= prog_sum r (r+n) (%m::nat. 
              alpha^(m-(1::nat)) * distance(xs 0, xs 1))"
apply (intro allI)
apply (induct_tac n)
apply (simp add: prog_sum_def)
apply (intro impI)
apply (simp)
apply (simp add: prog_sum_def)
apply (insert Banach_lm_contraction)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="r + na" in spec)
apply (simp)
done


lemma Banach_lm_ineq:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (r::nat) (n::nat).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i))) --> 
       distance(xs r, xs (r+n)) <= 
         distance (xs 0, f(xs 0)) * alpha^r / (1-alpha)"
apply (intro allI impI)
apply (insert Banach_lm_triangle)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="r" in spec)
apply (drule_tac x="n" in spec)

apply (insert Banach_lm_geo_prog_sum)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="r" in spec)
apply (drule_tac x="n" in spec)

apply (insert geo_prog_sum_infinite_div)
apply (drule_tac x="r" in spec)
apply (drule_tac x="r+n" in spec)
apply (drule_tac x="distance (xs 0, xs 1)" in spec)
apply (drule_tac x="alpha" in spec)

apply (simp add: contraction_alpha_def map_alpha_def positive_ms)
done

lemma Banach_lm_cauchy_lm_pos_distance:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (delta::real).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)) & 0 < delta 
      & 0 < distance (xs 0, f(xs 0)))
         --> (EX n. ALL m. distance (xs n, xs (n+m)) < delta)"
apply (intro allI impI)

apply (insert pow_convergence)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="delta*(1-alpha)/distance (xs 0, f(xs 0))" in spec)
apply (drule mp)

apply (simp add: ALL_real_divide_less_eq)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (simp add: ALL_real_mult_order)

apply (erule exE)
apply (rule_tac x="n" in exI)
apply (rule allI)

apply (insert Banach_lm_ineq)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m" in spec)
apply (simp)

apply (insert ALL_real_mult_div_commute)
apply (drule_tac x="alpha ^ n" in spec)
apply (drule_tac x="delta" in spec)
apply (drule_tac x="(1 - alpha)" in spec)
apply (drule_tac x="distance (xs 0, f (xs 0))" in spec)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (simp add: zero_le_power)
done

lemma Banach_lm_cauchy_lm_zero_distance:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (delta::real).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)) & 0 < delta 
      & distance (xs 0, f(xs 0))=0)
         --> (EX n. ALL m. distance (xs n, xs (n+m)) < delta)"
apply (intro allI impI)
apply (rule_tac x="0" in exI)
apply (rule allI)

apply (insert Banach_lm_ineq)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="0" in spec)
apply (drule_tac x="m" in spec)
apply (auto)
done


lemma Banach_lm_cauchy_lm:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (delta::real).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)) & 0 < delta)
         --> (EX n. ALL m. distance (xs n, xs (n+m)) < delta)"
apply (intro allI impI)
apply (insert positive_ms)
apply (drule_tac x="xs 0" in spec)
apply (drule_tac x="f(xs 0)" in spec)

apply (insert Banach_lm_cauchy_lm_zero_distance)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)

apply (insert Banach_lm_cauchy_lm_pos_distance)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta" in spec)

apply (case_tac "distance (xs 0, f (xs 0)) = 0")
apply (simp_all)
done

lemma Banach_lm_cauchy:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)))
        --> cauchy xs"
apply (intro allI impI)
apply (simp add: cauchy_def)
apply (intro allI impI)

apply (insert Banach_lm_cauchy_lm)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="delta/2" in spec)
apply (simp)

apply (erule exE)
apply (rule_tac x="n" in exI)
apply (intro allI impI)
apply (frule_tac x="i-n" in spec)
apply (drule_tac x="j-n" in spec)
apply (simp)

apply (insert triangle_inequality_ms)
apply (drule_tac x="xs i" in spec)
apply (drule_tac x="xs n" in spec)
apply (drule_tac x="xs j" in spec)

apply (insert symmetry_ms)
apply (drule_tac x="xs i" in spec)
apply (drule_tac x="xs n" in spec)
apply (auto)
done

(*** step 2 (converge to a fixpoint) ***)

lemma Banach_lm_converge:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)))
        --> (EX x. xs convergeTo x)"
apply (intro allI impI)
apply (insert Banach_lm_cauchy)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (simp add: comp_metirc_space)
done

lemma Banach_lm_fixpoint_lm:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (y::'a::cms) (eps::real).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)) 
      & xs convergeTo y & 0 < eps)
          --> distance(y, f y) < eps"
apply (intro allI impI)
apply (simp add: convergeTo_def)
apply (elim conjE)

apply (rotate_tac 2)
apply (drule_tac x="eps/2" in spec)
apply (simp)
apply (erule exE)
apply (rotate_tac -1)
apply (frule_tac x="n" in spec)
apply (drule_tac x="Suc n" in spec)
apply (simp)

apply (insert triangle_inequality_ms)
apply (drule_tac x="y" in spec)
apply (drule_tac x="xs (Suc n)" in spec)
apply (drule_tac x="f y" in spec)

apply (simp add: contraction_alpha_def map_alpha_def)
apply (elim conjE)

apply (rotate_tac -1)
apply (drule_tac x="xs n" in spec)
apply (drule_tac x="y" in spec)

apply (insert symmetry_ms)
apply (drule_tac x="xs n" in spec)
apply (drule_tac x="y" in spec)
apply (simp)

apply (insert ALL_le_than_one)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="distance (f (xs n), f y)" in spec)
apply (drule_tac x="distance (y, xs n)" in spec)

apply (simp add: positive_ms)
done


lemma Banach_lm_fixpoint:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (y::'a::cms).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)) 
      & xs convergeTo y) --> y = f y"
apply (intro allI impI)

apply (case_tac "0 < distance(y, f y)")

 apply (insert Banach_lm_fixpoint_lm)
 apply (drule_tac x="f" in spec)
 apply (drule_tac x="alpha" in spec)
 apply (drule_tac x="xs" in spec)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="distance (y, f y)" in spec)
 apply (simp)

apply (case_tac "0 = distance(y, f y)")
 apply (simp add: diagonal_ms)

(* distance(y, f y) < 0 *)
 apply (insert positive_ms)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="f y" in spec)
 apply (simp)
done

(*** step 3 (unique fixpoint) ***)

lemma Banach_lm_unique_lm:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) 
     (y::'a::cms) (z::'a::cms) (n::nat).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i)) 
      & y = f y & z = f z) --> distance(y, z) <= alpha * distance(y, z)"
apply (intro allI impI)
apply (simp add: contraction_alpha_def map_alpha_def)
apply (elim conjE)
apply (drule_tac x="y" in spec)
apply (drule_tac x="z" in spec)
apply (simp)
done

lemma Banach_lm_unique:
  "ALL f (alpha::real) (xs::'a::cms infinite_seq) (y::'a::cms) (z::'a::cms).
     (contraction_alpha f alpha & (ALL i. xs (Suc i) = f(xs i))
      & y = f y & z = f z) --> (y = z)"
apply (intro allI impI)

apply (case_tac "0 < distance(y, z)")
 apply (insert Banach_lm_unique_lm)
 apply (drule_tac x="f" in spec)
 apply (drule_tac x="alpha" in spec)
 apply (drule_tac x="xs" in spec)
 apply (drule_tac x="y" in spec)
 apply (drule_tac x="z" in spec)
 apply (drule_tac x="n" in spec)
 apply (simp add: contraction_alpha_def)

 apply (insert ALL_real_mult_less_cancel_if1)
 apply (drule_tac x="alpha" in spec)
 apply (drule_tac x="1" in spec)
 apply (drule_tac x="distance (y, z)" in spec)
 apply (simp)

apply (case_tac "0 = distance(y, z)")
 apply (simp add: diagonal_ms)

(* distance(y, Z) < 0 *)
apply (simp add: positive_ms real_less_le)
done


(*** final step (Banach Theory) ***)

theorem Banach:
  "ALL f (xs::'a::cms infinite_seq).
     (contraction f & (ALL i. xs (Suc i) = f(xs i)))
        --> (EX y. (xs convergeTo y 
                    & y = f y
                    & (ALL z. z = f z --> y = z)))"
apply (intro allI impI)
apply (simp add: contraction_def)
apply (elim conjE exE)
apply (insert Banach_lm_converge)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (simp)
apply (erule exE)
apply (rename_tac f xs alpha y)

apply (rule_tac x="y" in exI)
apply (simp)

apply (insert Banach_lm_fixpoint)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="y" in spec)
apply (simp)

apply (intro allI impI)
apply (insert Banach_lm_unique)
apply (drule_tac x="f" in spec)
apply (drule_tac x="alpha" in spec)
apply (drule_tac x="xs" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="z" in spec)
apply (simp)
done

end



